/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf.actuators;

import de.unibi.airobots.bonsai.actuators.BtlMemoryActuator;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.Type;
import de.unibi.airobots.btl.data.Type.NoSourceDocumentException;
import java.io.IOException;
import net.sf.xcf.ActiveMemory;
import net.sf.xcf.memory.DocumentNotFoundException;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;
import org.apache.log4j.Logger;

/**
 * A simple {@link Actuator} for BTL based types that uses their static
 * <code>fromDocument</code> method to extract the data. This allows this 
 * actuator to be used for every ActiveMemory that provides a BTL 
 * parsable result.
 * 
 * @param <T> BTL type that is inserted or replaced by this actuator.
 * 
 * @author lkettenb
 */
public class XcfBtlMemoryActuator<T extends Type> implements 
        BtlMemoryActuator<T>, ActiveMemoryReconnectable {
    
    /*
     * Error messages.
     */
    private static final String ERR_MEM_NOT_RECHABLE = 
            "Scene memory is not reachable.";
    private static final String ERR_OBJ_NOT_FOUND = 
            "Object can not be updated. " + 
            "Corresponding document not found in memory.";
    
    /**
     * Instance of the active memory that is used.
     */
    private ActiveMemory activeMemory = null;
    
    /**
     * Name of the active memory.
     */
    private String memoryName = "";
    
    /**
     * Lock for synchronized memory access.
     */
    private final Object memoryLock = new Object();
    
    /**
     * The log.
     */
    private Logger logger = Logger.getLogger(this.getClass());
    
    /**
     * Constructor.
     * 
     * @param memoryName Name of the active memory to work on.
     */
    public XcfBtlMemoryActuator(String memoryName) {
        this.memoryName = memoryName;
    }
    
    
    @Override
    public void insert(T object, Class<? extends Type> type) throws 
            IOException {
        if (activeMemory == null) {
            logger.error(ERR_MEM_NOT_RECHABLE);
        }
        try {
            activeMemory.insert(new XOPData(object.toDocument(type)));
        } catch (MemoryException e) {
            logger.error(ERR_MEM_NOT_RECHABLE);
            logger.debug(e);
            throw new IOException(e);
        }
    }
    
    @Override
    public void replace(T object) throws 
            IOException, NoSourceDocumentException {
        if (activeMemory == null) {
            logger.error(ERR_MEM_NOT_RECHABLE);
        } else {
            try {
                // copy source document
                Element source = new Element(object.getSourceDocument());

                // fill source document with new data
                object.fillInto(source);
                activeMemory.replace(new XOPData(new Document(source)));
            } catch (DocumentNotFoundException e) {
                logger.error(ERR_OBJ_NOT_FOUND);
                logger.debug(e);
                throw new IOException(e);
            } catch (MemoryException e) {
                logger.error(ERR_MEM_NOT_RECHABLE);
                logger.debug(e);
                throw new IOException(e);
            }
        }
    }
    
    
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }
    
    
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {
        synchronized (memoryLock) {
            this.activeMemory = activeMemory;
        }
    }
}
