/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf.sensors;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Attribute;
import nu.xom.Document;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.Type;

/**
 * A simple {@link Sensor} for BTL based types that uses their static
 * <code>fromDocument</code> method to extract the data. This allows this sensor
 * to be used for every ActiveMemory that provides a BTL parseable result.
 * 
 * @author lziegler
 * @param <T>
 *            btl type returned by this sensor
 */
public class XcfBtlMemorySensor<T extends Type> implements Sensor<T>,
        ActiveMemoryReconnectable {

    private Logger logger = Logger.getLogger(getClass());

    private Class<T> typeClass;
    private String memoryName = "";
    private Object memoryLock = new Object();
    private Object sharedDataLock = new Object();
    private ActiveMemory memory = null;

    private Method fromDocumentMethod = null;
    private Method getBaseTagMethod = null;

    /**
     * Namespace of the documents.
     */
    public static final String NAMESPACE_DBXML =
            "http://www.sleepycat.com/2002/dbxml";

    private List<SensorListener<T>> listeners =
            new LinkedList<SensorListener<T>>();

    private Map<Integer, T> objects = new HashMap<Integer, T>();
    private T lastObject = null;

    /**
     * Own listener used to intercept the memory for data changes.
     */
    private MemoryEventListener selfListener;

    /**
     * Constructor.
     * 
     * @param memoryName
     *            name of the active memory to get the data from
     * @param typeClass
     *            A {@link Class} object representing the expected instance's
     *            type.
     */
    public XcfBtlMemorySensor(Class<T> typeClass, String memoryName) {

        this.memoryName = memoryName;
        this.typeClass = typeClass;

        // find fromDocument method for the btl type
        try {
            fromDocumentMethod =
                    typeClass.getMethod("fromDocument", new Class[] {
                            Document.class, Class.class });
        } catch (SecurityException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        }

        // find getBaseTag method for the btl type
        try {
            getBaseTagMethod =
                    typeClass.getMethod("getBaseTag", new Class[] {});
        } catch (SecurityException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'getBaseTag' method in the given type "
                            + typeClass + ".", e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'getBaseTag' method in the given type "
                            + typeClass + ".", e);
        }

        Object rawResult;
        try {
            rawResult = getBaseTagMethod.invoke(null);
            if (rawResult == null) {
                throw new IllegalArgumentException(
                        "getBaseTag method has wrong signature "
                                + "or simply returned null.");
            }
        } catch (IllegalArgumentException e1) {
            throw new IllegalArgumentException(
                    "getBaseTag method has wrong signature.", e1);
        } catch (IllegalAccessException e1) {
            throw new IllegalArgumentException(
                    "getBaseTag can not be accessed.", e1);
        } catch (InvocationTargetException e1) {
            throw new IllegalArgumentException(
                    "getBaseTag throws an exception.", e1);
        }

        String baseTag = (String) rawResult;

        logger.debug("Create scene memory sensor for memoryName " + memoryName
                + " and xpath " + baseTag);

        selfListener =
                new MemoryEventAdapter(MemoryAction.ALL, new XPath(baseTag)) {

                    @Override
                    public void handleEvent(MemoryEvent e) {
                        handleMemoryEvent(e);
                    }
                };
    }

    private void handleMemoryEvent(MemoryEvent e) {
        try {

            logger.debug("received memory event");

            Document doc = e.getData().getDocument();
            if (doc == null) {
                throw new IOException("Document empty in XOPData: "
                        + e.getData());
            }

            // create BTL type
            T obj = safeExtractType(doc);

            // extract xcf id
            Attribute att =
                    doc.getRootElement().getAttribute("id", NAMESPACE_DBXML);
            if (att == null) {
                logger.error("Received Document has no xcf id");
                return;
            }
            int id = Integer.parseInt(att.getValue());

            synchronized (sharedDataLock) {

                MemoryAction a = e.getType();
                if (a == MemoryAction.INSERT || a == MemoryAction.REPLACE) {
                    objects.put(id, obj);
                } else if (e.getType() == MemoryAction.REMOVE) {
                    objects.remove(id);
                }
                lastObject = obj;

                for (SensorListener<T> l : listeners) {
                    l.newDataAvailable(obj);
                }
                sharedDataLock.notify();
            }

        } catch (NumberFormatException ex) {
            logger.error("Received Document has not a valid xcf id");
            logger.debug(ex);
        } catch (IOException ex) {
            logger.error("IO Exception while extracting type from document");
            logger.debug(ex);
        }
    }

    /**
     * Safely extracts the BTL type from the document found in the
     * {@link XOPData}.
     * 
     * @param data
     *            data containing the document
     * @return extracted BTL type
     * @throws IOException
     *             error extracting BTL type instance
     */
    @SuppressWarnings("unchecked")
    private T safeExtractType(Document doc) throws IOException {

        try {
            Object rawResult = fromDocumentMethod.invoke(null, doc, typeClass);
            if (rawResult == null) {
                throw new IOException(
                        "fromDocument method has wrong signature "
                                + "or simply returned null.");
            }
            return (T) rawResult;
        } catch (IllegalArgumentException e) {
            throw new IOException(e);
        } catch (IllegalAccessException e) {
            throw new IOException(e);
        } catch (InvocationTargetException e) {
            throw new IOException("Error parsing the document.", e);
        } catch (ExceptionInInitializerError e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {

        synchronized (memoryLock) {
            this.memory = activeMemory;

            try {
                memory.addListener(selfListener);
            } catch (MemoryException e) {
                logger.error(e);
            }
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public Class<T> getDataType() {
        return typeClass;
    }

    @Override
    public T readLast() throws IOException, InterruptedException {
        T o = null;
        synchronized (sharedDataLock) {
            o = lastObject;

            if (o == null) {
                sharedDataLock.wait();

                o = lastObject;
            }
        }
        return o;
    }

    @Override
    public T readLast(long timeout) throws IOException, InterruptedException {
        T o = null;
        synchronized (sharedDataLock) {
            o = lastObject;

            if (o == null) {
                sharedDataLock.wait(timeout);
                o = lastObject;
            }
        }
        return o;
    }

    public Collection<T> getAllCurrentObjects() {
        return objects.values();
    }

    @Override
    public boolean hasNext() {
        return lastObject != null;
    }

    @Override
    public void addSensorListener(SensorListener<T> listener) {
        listeners.add(listener);

    }

    @Override
    public void removeSensorListener(SensorListener<T> listener) {
        listeners.remove(listener);
    }
}
